"""Currently all moved to f311.filetypes.basic.misc"""

"""
Miscellanea of basic routines used to implement the file classes
"""

import re
import os
import shutil
import sys
import a99
import unicodedata


__all__ = ["SS_PLAIN", "SS_SUPERSCRIPT", "SS_ALL_SPECIAL", "SS_RAW",
           "greek_to_spdf", "spdf_to_greek", "state_to_str", "molconsts_to_system_str",
           "adjust_atomic_symbol", "str_to_elem_ioni", "parse_system_str",
           "split_molecules_description", "description_to_symbols", "symbols_to_formula",
           "iz_to_branch", "branch_to_iz", "iz_to_branch_alt", "branch_to_iz_alt",
           "branch_to_iz_alt2",
           ]

# **        ****                ******        ****                ******        ****
#   **    **    ******    ******      **    **    ******    ******      **    **    ******    ******
#     ****            ****              ****            ****              ****            ****
#
# Multiplicity- and spdf-related conversion and formatting routines

# # System styles
# Example: "A 2 SIGMA - X 2 PI"
SS_PLAIN = -1
# multiplicity as superscript, spdf as Greek letter name
SS_SUPERSCRIPT = 2
# multiplicity as superscript, spdf as Greek letter
SS_ALL_SPECIAL = 0
# Does not convert SPDF to string. Example: "A 2 0 - X 2 1"
SS_RAW = 1

__A = ["Sigma", "Pi", "Delta", "Phi"]

_SPDF_TO_GREEK = dict(zip(range(len(__A)), __A))
_GREEK_TO_SPDF = dict(zip(__A, range(len(__A))))

def greek_to_spdf(greek):
    """Converts Greek letter name (e.g., "sigma", "Sigma" (case **sensitive**)) to int value in [0, 1, 2, 3]"""
    try:
        ret = _GREEK_TO_SPDF[greek]
    except KeyError:
        raise ValueError("Invalid SPDF: '{}' (value must be in {})".format(greek, list(_SPDF_TO_GREEK.values())))
    return ret

def spdf_to_greek(number):
    """Converts int value in [0, 1, 2, 3] to the name of a Greek letter (all uppercase)"""
    try:
        return _SPDF_TO_GREEK[number]
    except KeyError:
        # "?" is the "zero-element"
        return "?"


def state_to_str(label, mult, spdf, style=SS_ALL_SPECIAL):
    """Compiles electronic state information into string

    Args:
        label: e.g., "A"
        mult: e.g., 2 (meaning doublet)
        spdf: e.g., 0 (meaning Sigma)

        style: rendering style: one of SS_*

    Returns:
        str
    """

    if style == SS_PLAIN:
        fmult = lambda x: x
        fspdf = lambda x:spdf_to_greek(x)
    elif style == SS_ALL_SPECIAL:
        fmult = lambda x: a99.int_to_superscript(x)
        fspdf = lambda x: a99.greek_to_unicode(spdf_to_greek(x).capitalize())
    elif style == SS_RAW:
        fmult = lambda x: x
        fspdf = lambda x: x
    elif style == SS_SUPERSCRIPT:
        fmult = lambda x: a99.int_to_superscript(x)
        fspdf = lambda x:spdf_to_greek(x)
    else:
        raise ValueError("Invalid style: {}".format(style))

    return "{} {} {}".format(label, fmult(int(mult)), fspdf(spdf))


def molconsts_to_system_str(molconsts, style=SS_ALL_SPECIAL):
    """Compiles electronic system information into string

    Args:
        molconsts: dict-like containing keys 'from_label', 'from_mult', 'from_spdf', 'to_label',
                    'to_mult', 'to_spdf'

        style: rendering style: one of SS_*

    Returns:
        str
    """
    return "{} - {}".format(
        state_to_str(molconsts["from_label"], molconsts["from_mult"], molconsts["from_spdf"], style),
        state_to_str(molconsts["to_label"], molconsts["to_mult"], molconsts["to_spdf"], style),
    )

# **        ****                ******        ****                ******        ****
#   **    **    ******    ******      **    **    ******    ******      **    **    ******    ******
#     ****            ****              ****            ****              ****            ****
#
# Conversion and formatting routines: Atomic symbols and formulas


def adjust_atomic_symbol(x):
    """Makes sure atomic symbol is right-aligned and upper case (PFANT convention)."""
    assert isinstance(x, str)
    return "%2s" % (x.strip().upper())


def str_to_elem_ioni(search):
    """str_to_elem_ioni("C1") --> (adjust_atomic_symbol("C"), 1)"""
    r = re.search("(\w+)(\d+)", search)
    if r is None:
        raise ValueError(
            "Invalid search string: '{}' (must be '<element><ioni>')".format(search))
    elem, ioni = r.groups()
    elem = adjust_atomic_symbol(elem)
    ioni = int(ioni)
    return elem, ioni


# Formulas that can be recognized in PFANT molecule header information ("titulo")
# This exists for backward compatibility. New molecules should have its symbols informed separately
#
# **Careful** If you add entried, make sure everything is in uppercase
_BUILTIN_FORMULAS = {
     "MGH": ["MG", "H"],
     "C2": ["C", "C"],
     "CN": ["C", "N"],
     "CH": ["C", "H"],
     "13CH": ["A", "H"],
     "12C16O": ["C", "O"],
     "NH": ["N", "H"],
     "OH": ["O", "H"],
     "FEH": ["FE", "H"],
     "TIO": ["TI", "O"],
     "CO": ["C", "O"],
}

__TRBLOCK = [lambda x: x, lambda x: int(x), lambda x: greek_to_spdf(x.capitalize())]
# lambda functions to convert the parsed bits, in sync with (from_label, ..., to_spdf)
_PSS_TRANSFORMS = __TRBLOCK * 2

def parse_system_str(string):
    """
    Parses "system string" --> (from_label, from_mult, from_spdf, to_label, to_mult, to_spdf)

    Args:
        string: see convention below

    **Convention** system string::

        "from_label from_mult from_spdf_greek - from_label from_mult from_spdf_greek"
        or
        "label mult spdf_greek" (assumed initial and final state are the same).

        **Note** (*spdf*) are Greek letter names, case-INsensitive;
                 extra "+"/"-" (for example, in "SIGMA+") is ignored

    System string examples::

        "[A 2 Sigma - X 2 Pi]"

        "[X 1 SIGMA+]"

    """

    expr = re.compile(
        "\[\s*([a-zA-Z])\s*(\d+)\s*([a-zA-Z0-9]+)[+-]{0,1}\s*-+\s*\s*([a-zA-Z])\s*(\d+)\s*([a-zA-Z0-9]+)[+-]{0,1}\s*\]")
    groups = expr.search(string)
    if groups is not None:
        # TODO this may be giving error
        _pieces = [groups[i] for i in range(1, 7)]
    else:
        # Initial and final state are the same. Example "12C16O INFRARED [X 1 SIGMA+]"
        expr = re.compile("\[\s*([a-zA-Z])\s*(\d+)\s*([a-zA-Z0-9]+)[+-]{0,1}\s*\]")

        groups = expr.search(string)
        if groups is not None:
            _pieces = [groups[i] for i in range(1, 4)] * 2

        if groups is None:
            raise ValueError("Could not understand str '{}'".format(string))


    pieces = [f(piece) for f, piece in zip(_PSS_TRANSFORMS, _pieces)]

    return pieces


def parse_label_mult_spdf(s):
    """Parses strings such as "A 2 SIGMA", "b3Sigmag-", "a3Piu" ...

    Returns: (label, multiplicity, spdf)

    Examples:
        'A 2 SIGMA' --> ('A', 2, 0)
        'b3Sigmag-' --> ('b', 3, 0)
        'a3Piu' --> ('a', 3, 1)

    Note: This routine is more complete than parse_system_str() (the latter cannot parse Greek letters with subsequent
    trailing letter (e.g. Piu)

    Note: I wrote this to parse file '12C12C_15000-17500_P-BR.bsyn' and similar, but later realized I just need the
          "label", so ended up not using this function, although it works.
    """

    s = s.strip()

    expr = re.compile("([a-zA-Z])\s*(\d+)\s*([a-zA-Z]+)")
    groups = expr.search(s)
    label, _mult, _spdf = [groups[i] for i in range(1, 4)]
    mult = int(_mult)
    _spdf_ = _spdf.capitalize()
    spdf = None
    for i, greek in enumerate(("Sigma", "Pi", "Delta", "Phi")):
        if _spdf_.startswith(greek):
            spdf = i
            break
    if spdf is None:
        raise ValueError(f"Invalid greek letter in '{s}'")
    return label, mult, spdf



def split_molecules_description(descr):
    """Breaks PFANT molecule description into "(name) (system) (optional notes)" --> (name, system, notes)

    System is identified by enclosing square brackets ("[", "]")
    """

    match = re.match("(.+?)\s*(\[.+\])\s*(.*)", descr)

    if match is None:
        raise ValueError("Invalid PFANT molecule description: '{}'".format(descr))

    return match.groups()


def description_to_symbols(descr):
    """Finds molecular formula in description text and returns separate symbols

    This routine is similar to pfantlib.f90:assign_symbols_by_formula()

    It is used when the symbols are not informed separately
    after the first "#" in the molecule 'titulo'

    Args:
        descr: string

    Returns:
        list of two elements, e.g., [" C", " O"]

    Example:

    >>> import pyfant
    >>> pyfant.description_to_symbols("12C16O INFRARED  X 1 SIGMA+,   version 15/oct/98")
    [' C', ' O']
    """

    descr = descr.upper()
    for formula, symbols in _BUILTIN_FORMULAS.items():
        # Searches for formula at start or after whitespace
        if re.search(r'(^|\W){}([^0-9A-Z]|$)'.format(formula), descr):
            return [adjust_atomic_symbol(x) for x in symbols]
    return None
    # raise RuntimeError("Could not find a valid formula in description '{}'".format(descr))


def symbols_to_formula(symbols):
    """Converts two symbols to a formula as recorded in the 'molecule' table

    Args:
        symbols: 2-element list as returned by f311.filetypes.basic.description_to_symbols()

    Returns:
        str: examples: 'MgH', 'C2', 'CH'

    Formulas in the 'molecule' table have no isotope information
    """
    rec = re.compile("[a-zA-Z]+")
    symbols_ = [rec.search(x).group().capitalize() for x in symbols]
    symbols_ = ["C" if x == "A" else x for x in symbols_]
    if symbols_[0] == symbols_[1]:
        return "{}{}".format(symbols_[0], "2")
    else:
        return "".join(symbols_)


# (20230725) THIS WAS ABANDONED TODO CLEANUP
#  def parse_label_mult_spdf(s):
#     """Parses strings such as "A 2 SIGMA", "b3Sigmag-", "a3Piu" ...
#
#     Returns: (label, multiplicity, spdf)
#
#     Note: This routine is more complete than parse_system_str() (the latter cannot parse Greek letters with subsequent
#     trailing letter (e.g. Piu)
#     """
#
#     s = s.strip()
#
#     expr = re.compile("([a-zA-Z])\s*(\d+)\s*([a-zA-Z]+)")
#     groups = expr.search(string)
#     if groups is not None:
#         _pieces = groups[1:4]
#     else:
#         # Initial and final state are the same. Example "12C16O INFRARED [X 1 SIGMA+]"
#         expr = re.compile("\[\s*([a-zA-Z])\s*(\d+)\s*([a-zA-Z0-9]+)[+-]{0,1}\s*\]")
#
#         groups = expr.search(string)
#         if groups is not None:
#             _pieces = [groups[i] for i in range(1, 4)] * 2
#
#         if groups is None:
#             raise ValueError("Could not understand str '{}'".format(string))
#
#
#     pieces = [f(piece) for f, piece in zip(_PSS_TRANSFORMS, _pieces)]
#
#     return pieces





# **        ****                ******        ****                ******        ****
#   **    **    ******    ******      **    **    ******    ******      **    **    ******    ******
#     ****            ****              ****            ****              ****            ****
#
# Conversion and formatting routines: Branch-related
#

_iz_to_branch_map = {"1": "P", "2": "Q", "3": "R", "4": "P1", "5": "Q1", "6": "R1", "7": "P2",
                     "8": "Q2", "9": "R2", "10": "P3", "11": "Q3", "12": "R3", }
_branch_to_iz_map = dict(((value, key) for key, value in _iz_to_branch_map.items()))


# alternative mapping
_iz_to_branch_map_alt = {"1": "P1","2": "P12","3": "P21","4": "P2","5": "Q1","6": "Q12",
                         "7": "Q21","8": "Q2","9 ": "R1","10": "R12","11": "R21","12": "R2",}
_branch_to_iz_map_alt = dict(((value, key) for key, value in _iz_to_branch_map_alt.items()))


# another alternative mapping
_iz_to_branch_map_alt2 = {"1": "P1","2": "Q1","3": "R1","4": "P2", "5": "Q2","6": "R2",
                          "7": "P3","8": "Q3","9 ": "R3"}
_branch_to_iz_map_alt2 = dict(((value, key) for key, value in _iz_to_branch_map_alt2.items()))



def iz_to_branch(iz):
    """Converts BLB's 'iz' code to string P/Q/R/P1. Inverse of branch_to_iz()

    (see pfantlib.90:read_molecules())

    **Note** iz-to-branch conversion cannot be trusted because different molecules use different
             conventions. To solve this in the future, new PFANT molecular lines .dat files will
             have branch saved as string.

    TODO note above is important
    """
    return _iz_to_branch_map[str(iz)]


def branch_to_iz(br):
    """Converts branch P/Q/R/P1, etc. into BLB's 'iz' code

    Args:
        br: str, [P/Q/R][/1/2/3], for example "P", "P1"

    Returns:
        str: from "1" to "12"
    """
    return _branch_to_iz_map[br]


def iz_to_branch_alt(iz):
    """Alternative mapping used by Bruno Castilho's CH. Inverse of branch_to_iz_alt()

    This is the mapping found in Bruno Castilho's work for molecule CH
    (directory ATMOS/wrk4/bruno/Mole/CH, check file, e.g., sja000.dat and source selech.f)
    """
    return _iz_to_branch_map_alt[str(iz)]


def branch_to_iz_alt(br):
    """
    Alternative mapping used by Bruno Castilho's CH

    Args:
        br: str, [P/Q/R][1/12/21/2], for example "P1", "Q21"

    Returns:
        str: from "1" to "12"

    This is the mapping found in Bruno Castilho's work for molecule CH
    (directory ATMOS/wrk4/bruno/Mole/CH, check file, e.g., sja000.dat and source selech.f)
    """
    return _branch_to_iz_map_alt[br]

def branch_to_iz_alt2(br):
    """
    Still another alternative mapping (for triplets)

    Args:
        br: str: P1/P2/P3/Q1/Q2/Q3/R1/R2/R3

    Returns:
        str: from "1" to "9"

    This is the mapping found in Bruno Castilho's work for molecule NH
    (directory ATMOS/wrk4/bruno/Mole/NH, check file, e.g., sj0000.dat)
    """
    return _branch_to_iz_map_alt2[br]



# -*- FilePy: FileConfigConvMol -*-
#
# @ Now @ 2017-11-27 18:17:21.158684
#

from f311.filetypes import ConfigDict

ccm = ConfigDict({
 'sourcename': 'Kurucz',
 'fn_output': '',
 'kurucz_iso': None,
 'kurucz_fn_input': '',
 'flag_hlf': False,
 'flag_normhlf': False,
 'flag_fcf': False,
 'flag_special_fcf': False,
 'flag_spinl': False,
 'flag_quiet': False
})



"""
Self-contained sub-library

  - **must not** import modules from outside this directory (self-contained)
  - should contain the most basic and general-purpose routines of the project
    (have in mind that it could make another API)
  - no project-specific hard-coded data
"""

from .errors import *
from .constants import *
from .conversion import *
from .misc import *
from .paths import *



"""Graphicsl interface for PFANT - single and multi modes"""

__all__ = ["XMulti"]

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import os.path
import shutil
from ._shared import *
from .a_XPFANT import *
import a99
import f311
import pyfant

WINDOW_WIDTH = 700
WINDOW_HEIGHT = 768


################################################################################
class XMulti(XPFANT):
    """
    Args:
      parent=None: nevermind
      file_main (optional)-- FileMain instance
    """

    def __init__(self, *args, **kwargs):
        ## State variables
        XPFANT.__init__(self, *args, **kwargs)

        # # Complementing previous configuration

        w = self.buttonSubmit.setText("&Submit single job")
        self.ae.loaded.connect(self.on_file_abonds_loaded)

        # # Multi tab

        tt = self.tabWidget
        tt.currentChanged.connect(self.on_tabWidget_current_changed)

        wm = self.multi_widget = QWidget()
        TAB_TEXT_4 = "Multi mode (Alt+&4)"
        tt.addTab(wm, TAB_TEXT_4)
        tt.setTabToolTip(tt.count()-1, DESCR_MULTI)

        # ## Layout with all elements in Multi tab
        l = QVBoxLayout()
        wm.setLayout(l)

        # ### Row with filename
        l1 = self.c923788 = QHBoxLayout()
        l.addLayout(l1)
        w = QLabel("<b>File:<b>")
        l1.addWidget(w)
        w = self.label_fn_abxfwhm = QLabel()
        l1.addWidget(w)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # ### Toolbar
        l1 = self.multiToolbarLayour = QHBoxLayout()
        l.addLayout(l1)
        w = self.buttonSubmit = QPushButton("&Submit multi-job")
        w.clicked.connect(self.on_run_multi)
        l1.addWidget(w)
        w = self.checkbox_multi_custom_id = QCheckBox("Custom multi-session directory")
        w.stateChanged.connect(self.on_checkbox_multi_custom_id_state_changed)
        l1.addWidget(w)
        w = self.lineEdit_multi_custom_id = QLineEdit()
        l1.addWidget(w)
        w.setFixedWidth(100)
        # s = self.spacer0 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # ### Editor for multi setup

        editor = self.multi_editor = pyfant.WFileAbXFwhm()
        editor.changed.connect(self.on_multi_changed)
        l.addWidget(editor)

        # ## Registers data in a_XMainAbonds lists to automatically perform Load/Save/Save as
        self.tab_texts.append(TAB_TEXT_4)
        self.flags_changed.append(False)
        self.save_as_texts.append("Save abundances X FWHM's configuration as")
        self.open_texts.append("Load abundandex X FWHM's file")
        self.clss.append(pyfant.FileAbXFwhm)
        self.editors.append(self.multi_editor)
        self.labels_fn.append(self.label_fn_abxfwhm)
        self.wilds.append("*.py")

        self.__update_lineEdit_multi_custom_id()
        # tt.setCurrentIndex(3)
        # ## Loads abxfwhm file
        if os.path.isfile(pyfant.FileAbXFwhm.default_filename):
            f = pyfant.FileAbXFwhm()
            f.load()
            self.multi_editor.load(f)
        # ## calls slot to perform cross-check between FileAbonds and FileAbXFwhm
        if self.ae.f:
            self.on_file_abonds_loaded()
        self._update_labels_fn()

        a99.snap_left(self, 720)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for Qt library signals

    def on_run_multi(self):
        import pyfant as pf
        errors = self._check_single_setup()

        if not self.multi_editor.f:
            errors.append("abundances X FWHM's configuration not set")
        else:
            # forces validation because validity is linked to FileAbonds managed in other tab
            self.multi_editor.f.validate()
            if not self.multi_editor.flag_valid:
                errors.append("error(s) in abundances X FWHM's configuration")

        if self.checkbox_multi_custom_id.isChecked():
            s = self.__get_multi_custom_session_id()
            if len(s) == 0:
                errors.append("Please inform custom session id.")
            elif len(errors) == 0: # will only offer to remove directory if everything is ok so far
                dirname = pyfant.get_custom_multisession_dirname(s)
                if os.path.isdir(dirname):
                    r = QMessageBox.question(self, "Directory exists",
                     "Directory '%s' already exists.\n\nWould you like to remove it?" % dirname,
                     QMessageBox.Yes|QMessageBox.No, QMessageBox.Yes)
                    if r == QMessageBox.Yes:
                        try:
                            shutil.rmtree(dirname)
                        except Exception as e:
                            errors.append(str(e))
                    else:
                        return

        if len(errors) == 0:
            try:
                self.setEnabled(False)
                self.__submit_multi()
                self._manager_form.show()
            except Exception as e:
                errors.append(str(e))
                a99.get_python_logger().exception("Cannot submit multi-job")
            finally:
                self.setEnabled(True)

        if len(errors) > 0:
            a99.show_error("Cannot submit multi-job:\n  - "+("\n  - ".join(errors)))

    def on_checkbox_multi_custom_id_state_changed(self):
        self.__update_lineEdit_multi_custom_id()

    def on_tabWidget_current_changed(self, index):
        self.__toggle_single_toolbar(index != 3)
        if index == 3:
            self.multi_editor.validate()

    def on_abonds_edited(self):
        """Overrides method to update multi_editor.file_abonds."""
        XPFANT.on_abonds_edited(self)
        self.multi_editor.file_abonds = self.ae.f

    def on_multi_changed(self):
        self._on_changed()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for signals emited by ftpyfant widgets

    def on_file_abonds_loaded(self):
        self.multi_editor.file_abonds = self.ae.f

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Protected methods

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Internals

    def __get_multi_custom_session_id(self):
        return str(self.lineEdit_multi_custom_id.text()).strip()

    def __submit_multi(self):
        import pyfant as pf
        r = pyfant.MultiRunnable(self.me.f, self.ae.f, self.oe.f, self.multi_editor.f)
        if self.checkbox_multi_custom_id.isChecked():
            custom_id = self.__get_multi_custom_session_id()
            if pyfant.get_custom_multisession_dirname(custom_id) == custom_id:
                # Understands that session dirname prefix must be cleared
                r.sid.id_maker.session_prefix_singular = ""
            r.sid.id = custom_id
        self._rm.add_runnables([r])

    def __update_lineEdit_multi_custom_id(self):
        self.lineEdit_multi_custom_id.setEnabled(self.checkbox_multi_custom_id.isChecked())

    def __toggle_single_toolbar(self, flag_enabled):
        layout = self.controlLayout
        # http://stackoverflow.com/questions/5150182/loop-over-widgets-in-pyqt-layout
        items = (layout.itemAt(i) for i in range(layout.count()))
        for item in items:
            w = item.widget()
            if w:
                w.setEnabled(flag_enabled)



__all__ = ["FileMollist"]

"""Represents file such as mollist.dat (2023)"""

import a99
from f311 import DataFile
import tabulate


@a99.froze_it
class FileMollist(DataFile):
    """
    List of PFANT molecular lines files
    """
    flag_collect = False  # too generic, left out of file type auto-detection
    default_filename = "mollist.dat"
    attrs = ["filenames"]

    def __init__(self):
        DataFile.__init__(self)

        # List of HmapRow objects
        self.filenames = []

    def __len__(self):
        return len(self.filenames)

    def _do_load(self, filename):
        """Loads from file."""

        with open(filename, "r") as h:
            for line in h:
                line = line.strip()
                if not line.startswith("#") and len(line) > 0:
                    self.filenames.append(line)

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            for r in self.filenames:
                a99.write_lf(h, r)



"""Molecular lines converter from Brooke2014 to PFANT format."""

__all__ = ["ConvBrooke2014"]

import a99, math
import airvacuumvald as avv
from .conv import *


class ConvBrooke2014(Conv):
    """Converts Brooke2014 [1] molecular lines data to PFANT "sets of lines"

    There is a working example conv_brooke2014_cn.py in examples directory

    References:
        [1] Brooke, James SA, et al. "Einstein a coefficients and oscillator strengths for the A2Π–X2Σ+(red)
            and B2Σ+–X2Σ+(violet) systems and rovibrational transitions in the X2Σ+ State of CN."
            The Astrophysical Journal Supplement Series 210.2 (2014): 23.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _make_sols(self, f):
        """Sets-of-lines maker for ConvBrooke2014 class

        Args:
            f: FileBrooke2014 instance
        """
        import pyfant
        assert isinstance(f, pyfant.FileBrooke2014)

        log = self.log

        log.n = len(f)
        if log.n == 0:
            raise RuntimeError("Zero lines found")

        if self.mode != ConvMode.HLF:
            deltak = self.molconsts.get_deltak()
            S2l = self.molconsts.get_S2l()

        STATEL = self.molconsts["from_label"]
        STATE2L = self.molconsts["to_label"]

        for i, line in enumerate(f.lines):
            try:
                if self.flag_filter_labels and not (line.eSl == STATEL and line.eS2l == STATE2L):
                    log.skip_reasons[f"Wrong system: {line.eSl}-{line.eS2l}"] += 1
                    continue

                vl = line.vl
                v2l = line.v2l
                Jl = line.Jl
                J2l = line.J2l
                branch = line.branch
                nu = f.nu_obs_or_calc(i)
                lambda_ = avv.vacuum_to_air(1e8/nu)

                if nu < 0:
                    log.skip_reasons["Negative wavenumber"] += 1
                    continue

                if self.mode == ConvMode.HLF:
                    sj = self.get_hlf(vl, v2l, J2l, branch)
                else:
                    A = line.A
                    sj = self.get_sj_einstein(A, Jl, J2l, S2l, deltak, nu)

                self.append_line2(vl, v2l, lambda_, sj, J2l, branch)

            except Exception as e:
                reason = a99.str_exc(e)
                log.skip_reasons[reason] += 1
                msg = "#{}{} line: {}".format(i+1, a99.ordinal_suffix(i+1), reason)
                log.errors.append(msg)
                if not self.flag_quiet:
                    a99.get_python_logger().exception(msg)


# === SAMPLES ==========================================================================================================
# hapidata["header"]
# ==============
# {'table_type': 'column-fixed',
#  'size_in_bytes': -1,
#  'table_name': '###',
#  'number_of_rows': 253695,
#  'order': ['molec_id',
#   'local_iso_id',
#   'nu',
#   'sw',
#   'a',
#   'gamma_air',
#   'gamma_self',
#   'elower',
#   'n_air',
#   'delta_air',
#   'global_upper_quanta',
#   'global_lower_quanta',
#   'local_upper_quanta',
#   'local_lower_quanta',
#   'ierr',
#   'iref',
#   'line_mixing_flag',
#   'gp',
#   'gpp'],
#  'format': {'a': '%10.3E',
#   'gamma_air': '%5.4f',
#   'gp': '%7.1f',
#   'local_iso_id': '%1d',
#   'molec_id': '%2d',
#   'sw': '%10.3E',
#   'local_lower_quanta': '%15s',
#   'local_upper_quanta': '%15s',
#   'gpp': '%7.1f',
#   'elower': '%10.4f',
#   'n_air': '%4.2f',
#   'delta_air': '%8.6f',
#   'global_upper_quanta': '%15s',
#   'iref': '%12s',
#   'line_mixing_flag': '%1s',
#   'ierr': '%6s',
#   'nu': '%12.6f',
#   'gamma_self': '%5.3f',
#   'global_lower_quanta': '%15s'},
#  'default': {'a': 0.0,
#   'gamma_air': 0.0,
#   'gp': 'FFF',
#   'local_iso_id': 0,
#   'molec_id': 0,
#   'sw': 0.0,
#   'local_lower_quanta': '000',
#   'local_upper_quanta': '000',
#   'gpp': 'FFF',
#   'elower': 0.0,
#   'n_air': 0.0,
#   'delta_air': 0.0,
#   'global_upper_quanta': '000',
#   'iref': 'EEE',
#   'line_mixing_flag': 'EEE',
#   'ierr': 'EEE',
#   'nu': 0.0,
#   'gamma_self': 0.0,
#   'global_lower_quanta': '000'},
#  'description': {'a': 'Einstein A-coefficient in s-1',
#   'gamma_air': 'Air-broadened Lorentzian half-width at half-maximum at p = 1 atm and T = 296 K',
#   'gp': 'Upper state degeneracy',
#   'local_iso_id': 'Integer ID of a particular Isotopologue, unique only to a given molecule, in order or abundance (1 = most abundant)',
#   'molec_id': 'The Brooke2014 integer ID for this molecule in all its isotopologue forms',
#   'sw': 'Line intensity, multiplied by isotopologue abundance, at T = 296 K',
#   'local_lower_quanta': 'Rotational, hyperfine and other quantum numbers and labels for the lower state of a transition',
#   'local_upper_quanta': 'Rotational, hyperfine and other quantum numbers and labels for the upper state of a transition',
#   'gpp': 'Lower state degeneracy',
#   'elower': 'Lower-state energy',
#   'n_air': 'Temperature exponent for the air-broadened HWHM',
#   'delta_air': 'Pressure shift induced by air, referred to p=1 atm',
#   'global_upper_quanta': 'Electronic and vibrational quantum numbers and labels for the upper state of a transition',
#   'iref': 'Ordered list of reference identifiers for transition parameters',
#   'line_mixing_flag': 'A flag indicating the presence of additional data and code relating to line-mixing',
#   'ierr': 'Ordered list of indices corresponding to uncertainty estimates of transition parameters',
#   'nu': 'Transition wavenumber',
#   'gamma_self': 'Self-broadened HWHM at 1 atm pressure and 296 K',
#   'global_lower_quanta': 'Electronic and vibrational quantum numbers and labels for the lower state of a transition'}}

# Sample lines, although I won't have to parse them because hapi does that
# ========================================================================
# 55    2.248764 2.700-164 8.704E-07.07970.08664763.01990.76-.000265             41             41                    R  0      457665 5 8 2 2 1 7     6.0    2.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# 55    2.279785 6.503-162 8.685E-07.07970.08663631.47680.76-.000265             40             40                    R  0      457665 5 8 2 2 1 7     6.0    2.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# 56    2.288013 3.667-166 9.362E-07.07970.08665317.82150.76-.000265             41             41                    R  0      457665 5 8 2 2 1 7    36.0   12.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# 55    2.310683 1.737-159 8.638E-07.07970.08662478.03720.76-.000265             39             39                    R  0      457665 5 8 2 2 1 7     6.0    2.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# 56    2.320259 8.961-164 9.357E-07.07970.08664183.55380.76-.000265             40             40                    R  0      457665 5 8 2 2 1 7    36.0   12.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# 53    2.325385 1.473-164 1.002E-06.07970.08665843.01970.76-.000265             41             41                    R  0      457665 5 8 2 2 1 7     3.0    1.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# 52    2.331616 5.471-164 1.014E-06.07970.08665930.07650.76-.000265             41             41                    R  0      457665 5 8 2 2 1 7     6.0    2.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800


#=======================================================================================================================
# (20230503) Old code from 2017, didn't delete yet, maybe todo cleanup
# """
# Brooke2014-specific conversion
#
# References:
#   [1] Rothman, Laurence S., et al. "The Brooke2014 2004 molecular spectroscopic database."
#       Journal of Quantitative Spectroscopy and Radiative Transfer 96.2 (2005): 139-204.
#   [2] extraeBrooke2014.f (First version: Jorge Melendez-Moreno, January 1999)
# """
#
# # TODO not working
#
#
# from .convlog import *
# from collections import OrderedDict
# import airvacuumvald as avv
# import pyfant
#
#
# __all__ = ["Brooke2014_to_sols"]
#
#
# _OH_BRANCH_MAP_OH = {"QQ": "Q", "PP": "P", "RR": "R"}
#
#
#
# def _parse_local_lower_quanta_OH(Q2l):
#     """
#     Parses the "Lower-state 'local' quanta" for 'OH' (see [1], Table 4, group 2, observation 'b')
#
#     Args:
#         Q2l: 15-character string containing the lower-state 'local' quanta information of molecular
#         line. Example: " PP 12.5ff     "
#
#     Returns: (Br, J2l), where
#         Br: branch ("P"/"Q"/"R")
#         J2l: inferior "quantum number associated with the total angular momentum excluding nuclear
#              spin" [1]
#
#     **Note** will raise if cannot determine Br (branch)
#     """
#     brinfo = Q2l[1:3]
#     Br = _OH_BRANCH_MAP_OH.get(brinfo)
#     if Br is None:
#         raise ValueError("Cannot derive branch from '{}' ((local lower quanta) = '{}')".
#                          format(brinfo, Q2l))
#     J2l = float(Q2l[3:8])
#     return Br, J2l
#
# def _parse_local_lower_quanta_group2(Q2l):
#     """
#     Parses the "lower-state 'local' quanta" for molecules in [1], Table 4, group 2
#
#     Args, Returns: see _parse_local_lower_quanta_OH()
#     """
#
#     Br = Q2l[5:6]
#     J2l = float(Q2l[6:9])
#     return Br, J2l
#
#
# def _parse_local_lower_quanta_group6(Q2l):
#     """[...] [1], Table 4, group 6"""
#
#     Br = Q2l[3:4]
#     J2l = float(Q2l[4:9])
#     return Br, J2l
#
#
# def _parse_global_quanta_class1(V):
#     """
#     Parses the [upper/lower] "global quanta" ([1], Table 3, class 1
#
#     Args:
#         V: 15-character string containing the 'global' quanta information of molecular
#            line. Example: "       X3/2   8"
#
#     Returns: v1: "quantum number associated with the first normal mode of vibration" [1]
#     """
#     return int(V[13:15])
#
#
# def _parse_global_quanta_class2(V):
#     """[...] [1], Table 3, class 2"""
#     return int(V[13:15])
#
# def _parse_global_quanta_class3(V):
#     return int(V[13:15])
#
#
# __GROUP_MAP = (
#     (("OH",),
#      _parse_local_lower_quanta_OH),
#     (("CO2", "N2O", "CO", "HF", "HCl", "HBr", "HI", "OCS", "N2", "HCN", "C2H2", "NO+"),
#      _parse_local_lower_quanta_group2),
#     (("NO", "ClO"),
#      _parse_local_lower_quanta_group6),
# )
# _GROUP_DICT = dict(sum([[(mol, callable_) for mol in mols] for mols, callable_ in __GROUP_MAP], []))
#
# __CLASS_MAP = (
#     (("CO", "HF", "HCl", "HBr", "HI", "N2", "NO+"), _parse_global_quanta_class1),
#     (("O2",), _parse_global_quanta_class2),
#     (("NO", "OH", "ClO"), _parse_global_quanta_class3),
# )
# _CLASS_DICT = dict(sum([[(mol, callable_) for mol in mols] for mols, callable_ in __CLASS_MAP], []))
#
#
# def Brooke2014_to_sols(molconsts, lines, qgbd_calculator):
#     """
#     Converts Brooke2014 molecular lines data to PFANT "sets of lines"
#
#     Args:
#         molconsts: a dict-like object combining field values from tables 'molecule', 'state',
#                     'pfantmol', and 'system' from a FileMolDB database
#         lines: item from rom hapi.LOCAL_TABLE_CACHE (a dictionary)
#         qgbd_calculator: callable that can calculate "qv", "gv", "bv", "dv",
#                          e.g., calc_qbdg_tio_like()
#
#     Returns: (a list of ftpyfant.SetOfLines objects, a MolConversionLog object)
#     """
#
#     def append_error(msg):
#         log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))
#
#     # C     BAND (v',v'')=(VL,V2L)
#     # C     WN: vacuum wavenumber   WL : air wavelength
#     # C     J2L: lower state angular momentum number
#     # C     iso: isotope/ 26: 12C16O, 36: 13C16O, 27: 12C17O, 28: 12C18O
#     # C     ramo: branch, for P: ID=1, for R: ID=2
#     # C     name: file name/ coisovLv2L
#     # C     HL: Honl-London factor
#     # C     FR: oscillator strength
#
#     header = lines["header"]
#     data = lines["data"]
#
#     S = molconsts.get_S2l()
#     DELTAK = molconsts.get_deltak()
#     print(f"DELTAK = {DELTAK}")
#     sys.exit()
#     formula = molconsts["formula"]
#
#     n = header["number_of_rows"]
#     log = MolConversionLog(n)
#     sols = OrderedDict()  # one item per (vl, v2l) pair
#
#     try:
#         f_class = _CLASS_DICT[formula]
#     except KeyError:
#         log.errors.append("Cannot determine Table 3 'class' (Rothman et al. 2005) for formula '{}'".
#                           format(formula))
#         log.flag_ok = False
#
#     try:
#         f_group = _GROUP_DICT[formula]
#     except KeyError:
#         log.errors.append("Cannot determine Table 4 'group' (Rothman et al. 2005) for formula '{}'".
#                           format(formula))
#         log.flag_ok = False
#
#     if not log.flag_ok:
#         return [], log
#
#     for i in range(n):
#
#         try:
#             nu = data["nu"][i]
#             wl = avv.vacuum_to_air(1e8/nu)
#             Br, J2l = f_group(data["local_lower_quanta"][i])
#             #  ASSUMING SINGLET!!!
#             Jl = ph.singlet.quanta_to_branch(Br, J2l)
#             V = f_class(data["global_upper_quanta"][i])
#             V_ = f_class(data["global_lower_quanta"][i])
#             A = data["a"][i]
#
#             # A seguir normalizacion de factor de Honl-london (HLN)
#             Normaliza = 1/((2.0*J2l+1)*(2.0*S+1)*(2.0-DELTAK))
#
#             # A seguir teremos a forca de oscilador
#             # mas q sera normalizada segundo o programa da Beatriz
#             gf = Normaliza*1.499*(2*Jl+1)*A/(nu**2)
#
#             J2l_pfant = int(J2l)  # ojo, estamos colocando J2L-0.5! TODO ask BLB: we write J2l or J2l-.5? (20171114 I think this is wrong, molecules.dat has decimal values)
#         except Exception as e:
#             log.errors.append("#{}{} line: {}".format(i+1, a99.ordinal_suffix(i+1), a99.str_exc(e)))
#             continue
#
#         sol_key = (V, V_)  # (v', v'') transition (v_sup, v_inf)
#         if sol_key not in sols:
#             qgbd = qgbd_calculator(molconsts, sol_key[1])
#             qqv = qgbd["qv"]
#             ggv = qgbd["gv"]
#             bbv = qgbd["bv"]
#             ddv = qgbd["dv"]
#
#             sols[sol_key] = pyfant.SetOfLines(V, V_, qqv, ggv, bbv, ddv, 1.)
#         sol = sols[sol_key]
#         sol.append_line(wl, gf, J2l_pfant, Br)
#
#     return (list(sols.values()), log)



"""
Represents SQLite database of molecular constants
"""

import a99
from f311 import FileSQLiteDB
import tabulate
import re


__all__ = ["FileMolDB"]


class FileMolDB(FileSQLiteDB):
    """Database of Molecular Constants"""
    default_filename = "moldb.sqlite"
    editors = ["convmol.py", "moldbed.py"]
    gui_info = {
        "molecule": {
            "formula": [None, "formula of molecule, <i>e.g.</i>, 'OH'"],
        },
        "pfantmol": {
            'description': [None, "free text"],
            "fe": [None, "oscillator strength"],
            "do": [None, "dissociation constant (eV)"],
            "am": [None, "mass of first element"],
            "bm": [None, "mass of second element"],
            "ua": [None, "value of partition function for first element"],
            "ub": [None, "value of partition function for second element"],
            "te": [None, "electronic term"],
            "cro": [None, "delta Kronecker (1: Sigma-Sigma transitions; 0: non-Sigma transitions)"],
        },
        "state": {
        "omega_e": ["ω<sub>e</sub>", "vibrational constant – first term (cm<sup>-1</sup>)"],
        "omega_ex_e": ["ω<sub>e</sub>x<sub>e</sub>", "vibrational constant – second term (cm<sup>-1</sup>)"],
        "omega_ey_e": ["ω<sub>e</sub>y<sub>e</sub>", " vibrational constant – third term (cm<sup>-1</sup>)"],
        "B_e": ["B<sub>e</sub>", "rotational constant in equilibrium position (cm<sup>-1</sup>)"],
        "alpha_e": ["α<sub>e</sub>", "rotational constant – first term (cm<sup>-1</sup>)"],
        "D_e": ["D<sub>e</sub>", "centrifugal distortion constant (cm<sup>-1</sup>)"],
        "beta_e": ["β<sub>e</sub>", "rotational constant – first term, centrifugal force (cm<sup>-1</sup>)"],
        "A": [None, "Coupling counstant (cm<sup>-1</sup>)"],
          },
        "system": {
            "from_label": ["State'", None],
            "from_spdf": ["Λ'", "0/1/2/3 meaning Σ/Π/Δ/Φ"],
            "to_label": ['State"', None],
            "to_spdf": ['Λ"', "0/1/2/3 meaning Σ/Π/Δ/Φ"]
        }

    }

    def get_molconsts(self, system_str):
        """
        Parses string and return MolConsts object.

        Args:
            system_str: system description as found in moldbed.py --> "System description" box. For example,
                        "MgH [A 2 Pi - X 2 Sigma]". Also returned by MolConsts.get_str()
        """
        import pyfant
        ret = pyfant.MolConsts()
        ret.populate_all_using_str(self, system_str)
        return ret

    def query_molecule(self, **kwargs):
        """Convenience function to query 'molecule' table

        Args:
            **kwargs: filter fieldname=value pairs to be passed to WHERE clause

        Returns: sqlite3 cursor
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select * from molecule{} order by formula""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def query_state(self, **kwargs):
        """Convenience function to query 'state' table (joins with table molecule)

        Args, Returns: see query_molecule
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select molecule.formula, state.* from state
                 join molecule on state.id_molecule = molecule.id{}""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def query_pfantmol(self, **kwargs):
        """Convenience function to query 'pfantmol' table (joins with table molecule)

        Args, Returns: see query_molecule
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select * from pfantmol{}""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def query_system(self, **kwargs):
        """Convenience function to query 'system' table (joins with table molecule)

        Args, Returns: see query_molecule
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select * from system{}""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def query_fcf(self, **kwargs):
        """Convenience function to query 'fcf' table

        Args, Returns: see query_molecule
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select * from fcf{} order by vl, v2l""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def test_query_state(self):
        """
        Test function

        Example:

        >>> f = FileMolDB()
        >>> f.init_default()
        >>> conn = f.get_conn()
        >>> cursor = conn.execute("select * from state where id = 10")
        >>> row = cursor.fetchone()
        >>> print(row)
        MyDBRow([('id', 10), ('id_molecule', 2), ('State', 'F ⁱPi_u'), ('T_e', 75456.9), ('omega_e', 1557.5), ('omega_ex_e', None), ('omega_ey_e', None), ('B_e', 1.645), ('alpha_e', 0.019), ('gamma_e', None), ('D_e', 6e-06), ('beta_e', None), ('r_e', 1.307), ('Trans', 'F ← X R'), ('nu_00', 74532.9)])
        """

    def print_states(self, **kwargs):
        """
        Prints states table in console

        Args:
            **kwargs: arguments passed to query_state()

        Example:

        >>> f = FileMolDB()
        >>> f.init_default()
        >>> f.print_states(formula="OH")
        """
        r = self.query_state(**kwargs)
        data, header0 = a99.cursor_to_data_header(r)

        # ti = a99.get_table_info(_ALIAS, "state")
        # header = [(ti[name]["caption"] or name) if ti.get(name) else name for name in header0]

        header = header0

        print(tabulate.tabulate(data, header))

    def get_transition_dict(self):
        """
        Generates a dictionary where (molecule, state_from, state_to) can be searched to retrieve state rows

        """
        rm = self.query_molecule()
        ret = {}
        for row_molecule in rm:
            rs = self.query_state(id_molecule=row_molecule["id"])
            for row_state in rs:
                row_state.None_to_zero()

                trans_ = row_state["Trans"]
                if trans_ is None or trans_ == 0:
                    continue
                keys = self._formula_trans_to_tuples(row_molecule["formula"], trans_)

                if keys is None:
                    continue
                for key in keys:
                    ret[key] = row_state
        return ret

    def get_fcf_dict(self, id_system):
        """Returns a dictionary indexed by (vl, v2l) and FCFs as values"""

        _ret = self.query_fcf(id_system=id_system).fetchall()
        ret = {}
        for r in _ret:
            ret[(r["vl"], r["v2l"])] = r["value"]
        return ret

    def find_id_system(self, molconsts):
        """Returns id_system or None"""

        ff = ["from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf"]
        one = self.get_conn().execute("select id from system where id_molecule = ? and "+
                           " and ".join(["{} = ?".format(x) for x in ff]),
                           [molconsts["id_molecule"]]+[molconsts[x] for x in ff]).fetchone()
        if one is None:
            return None
        return one["id"]



    def insert_system_if_does_not_exist(self, molconsts, notes=""):
        """
        Inserts system if does not exist yet

        Args:
            molconsts: a MolConsts
            notes: table system.notes value, in case of new record
        """
        ff = ["id_molecule", "from_label", "from_mult", "from_spdf", "to_label", "to_mult",
              "to_spdf"]

        molconsts._i_need(ff)  # stay friendly, MolConsts

        id_ = self.find_id_system(molconsts)

        if id_ is not None:
            return id_


        values = [molconsts[x] for x in ff]+[notes]
        ff.append("notes")
        conn = self.get_conn()
        cursor = conn.execute("""insert into system ({}) values ({})""".format(", ".join(ff),
            ", ".join(["?"]*len(ff))), values)
        id_ = cursor.lastrowid
        return id_


    def _create_schema(self):
        conn = self.get_conn()
        c = conn.cursor()
        c.execute("""create table molecule (id integer primary key,
                                            formula text unique,
                                            name text
                                           )""")


        # I decided that I won't try to track down all the Physics to determine which of the
        # molecule-wide "constants" in PFANT molecular lines file (e.g. 'molecules.dat') are
        # dependent on the molecule, a (vl, v2l) transition thereof etc.
        #
        # Instead, it seems to be historically sensible to reproduce the structure found in that
        # file, i.e., create a table with the same fields found in the molecule header of the PFANT
        # molecular lines file (e.g. 'molecules.dat'), **no less, no more**
        #
        # **Note** 'symbol_a', 'symbol_b' matches element symbols found in 'dissoc.dat'
        #
        c.execute("""create table pfantmol (id integer primary key,
                                            id_system integer,
                                            description text,
                                            fe real,
                                            do real,
                                            am real,
                                            bm real,
                                            ua real,
                                            ub real,
                                            te real,
                                            cro real,
                                            notes text
                                           )""")

        # This "state" information comes from NIST Chemistry Web Book.
        #
        # The field names are exactly as extracted from the header rows of the tables of diatomic
        # molecular constants in that book
        #
        c.execute("""create table state (id integer primary key,
                                         id_molecule integer,
                                         State text,
                                         T_e real,
                                         omega_e real,
                                         omega_ex_e real,
                                         omega_ey_e real,
                                         B_e real,
                                         alpha_e real,
                                         gamma_e real,
                                         D_e real,
                                         beta_e real,
                                         r_e real,
                                         Trans text,
                                         nu_00 real,
                                         A real,
                                         label text,
                                         mult integer,
                                         spdf integer,                                         
                                         notes text
                                        )""")
                                         # s_label text,
                                         # s_multiplicity integer,
                                         # s_spdf integer,
                                         # s_parity text


        # A     2            Sigma
        # label multiplicity spdf
        #
        # spdf:
        #     Sigma: 0
        #     Pi: 1
        #     Delta: 2
        #     Phi: 3
        c.execute("""create table system
                     (id integer primary key,
                      id_molecule integer,
                      from_label text,
                      from_mult integer,
                      from_spdf integer,
                      to_label text,
                      to_mult integer,
                      to_spdf integer,
                      notes text
                      )""")

        # Franck-Condon Factors table
        #
        # Something that I noticed from Bruno Castilho's directory at ATMOS/wrk4/Mole: both CH and
        # 13CH use the same Franck-Condon factors (files 'sjalist.txt', 'sjblist.txt' etc).
        # Therefore, the 'fcf' table below have a N-1 relation to table 'system', not to table
        # 'pfantmol' (the latter is where there is a distinction between 13CH and CH)
        #
        c.execute("""create table fcf (id integer primary key,
                                       id_system integer,
                                       vl integer,
                                       v2l integer,
                                       value real,
                                       notes text
                                       )""")

        conn.commit()
        conn.close()

    @staticmethod
    def _formula_trans_to_tuples(formula, trans):
        """Generates several tuples (formula, state_from, state_to)

        Args:
            formula: chemical formula, such as "TiO". This must be the same formula found in a
                     the 'molecule' table
            trans: string such as "A ↔ X R", "D ← X R", "B → A R"

        Returns: list of tuples [(formula, state_from, state_to), ...]
                 Example:  [('TiO', 'A', 'B'), ('TiO', 'R', 'B')]

        """

        try:
            lr = [re.split(" ", x.strip()) for x in re.split("[↔←→]", trans)]
            if "→" in trans:
                from_ = lr[0]
                to = lr[1]
            elif "←" in trans:
                to = lr[0]
                from_ = lr[1]
            else:
                from_ = to = lr[0]+lr[1]

            pairs = [(f, t) for t in to for f in from_ if t != f]

            ret = [(formula,)+pair for pair in pairs]
        except IndexError:
            ret = None

        return ret




"""Parses Brooke 2014 molecular line list"""

__all__ = ["FileBrooke2014"]

import a99, math
from f311 import DataFile
from dataclasses import dataclass

@dataclass
class FileBrooke2014Line:
    # Upper electronic state
    eSl: str
    # Lower electronic state
    eS2l: str
    # Upper state vibrational level
    vl: int
    # Lower state vibrational level
    v2l: int
    # Upper state J level
    Jl: float
    # Lower state J level
    J2l: float
    # Upper state parity
    pl: str
    # Lower state parity
    p2l: str
    # observed wavenumber in cm^-1
    nu_obs: float
    # calculated wavenumber in cm^-1
    nu_calc: float
    # Einstein A value
    A: float
    # Transition description
    Des: str
    # Branch taken from description (above)
    branch: str

    def nu_obs_or_calc(self):
        nu = self.nu_obs
        if math.isnan(nu):
            nu = self.nu_calc
        return nu

@a99.froze_it
class FileBrooke2014(DataFile):
    """
    Brooke 2014 molecular lines file [1]

    Model file CN1214-Brookeetal-2014-list.txt provided by Bertrand Plez [2]. This file contains an introductory
    descriptive text. This class is prepared to load files either with or without this introductory text.

    References:
        [1] Brooke, James SA, et al. "Einstein a coefficients and oscillator strengths for the A2Π–X2Σ+(red)
            and B2Σ+–X2Σ+(violet) systems and rovibrational transitions in the X2Σ+ State of CN."
            The Astrophysical Journal Supplement Series 210.2 (2014): 23.

        [2] https://www.lupm.in2p3.fr/users/plez/
    """

    def __init__(self):
        DataFile.__init__(self)
        self.lines = []

    def __len__(self):
        return len(self.lines)

    def nu_obs_or_calc(self, i):
        """Returns observed or calculated wavenumber.

        Preference is observed, but if not present, returns calculated."""

        return self.lines[i].nu_obs_or_calc()

    def _do_load(self, filename):
        with open(filename, "r") as h:
            state = "line"
            k = -1
            for i, s in enumerate(h):
                try:
                    if i == 0 and s[:5] == "Title":
                        state = "exp_note4"
                    if state == "line":
                        k += 1

                        try:
                            nu_obs = float(s[38:49])
                        except ValueError:
                            nu_obs = float("nan")
                        Des = s[107:118]

                        line = FileBrooke2014Line(nu_obs = nu_obs,
                                                  eSl=s[0:1],
                                                  eS2l=s[2:3],
                                                  vl=int(s[4:6]),
                                                  v2l=int(s[7:9]),
                                                  Jl=float(s[10:15]),
                                                  J2l=float(s[16:21]),
                                                  pl=s[26:27],
                                                  p2l=s[28:29],
                                                  nu_calc=float(s[50:60]),
                                                  A=float(s[81:93]),
                                                  Des=Des,
                                                  branch=Des[1:Des.index("(")],
                                                  )
                        self.lines.append(line)

                    elif state == "exp_note4":
                        if s[:8] == "Note (4)":
                            state = "exp_dashes"
                    elif state == "exp_dashes":
                        if s[:8] == "--------":
                            state = "line"

                except Exception as e:
                    raise RuntimeError("Error around %d%s row of file '%s': \"%s\""%
                                       (i+1, a99.ordinal_suffix(i+1), filename, a99.str_exc(e))) from e


"""
Title: Einstein A Coefficients and Oscillator Strengths for the 
       A^2^{Pi}-X^2^{Sigma}^+^ (red) and B^2^{Sigma}^+^-X^2^{Sigma}^+^ (violet)
       Systems and Rovibrational Transitions in the X^2^{Sigma}^+^ State of CN
Authors: Brooke J.S.A., Ram R.S., Western C.M., Li G.,
         Schwenke D.W., Bernath P.F.
Table: List of observed and calculated positions and calculated intensities of 
       the CN A^2^{Pi}-X^2^{Sigma}^+^ and B^2^{Sigma}^+^-X^2^{Sigma}^+^ systems,
       and X^2^{Sigma}^+^ state rovibrational transitions
================================================================================
Byte-by-byte Description of file: apjs489210t4_mrt.txt
--------------------------------------------------------------------------------
   Bytes Format Units Label  Explanations
--------------------------------------------------------------------------------
       1 A1     ---   eS'    Upper electronic state
       3 A1     ---   eS''   Lower electronic state
   5-  6 I2     ---   v'     Upper state vibrational level
   8-  9 I2     ---   v''    Lower state vibrational level
  11- 15 F5.1   ---   J'     Upper state J level
  17- 21 F5.1   ---   J''    Lower state J level
      23 I1     ---   F'     Upper state F level (1)
      25 I1     ---   F''    Lower state F level (1)
      27 A1     ---   p'     Upper state parity
      29 A1     ---   p''    Lower state parity
  31- 33 A3     ---   N'     Upper state N level
  35- 37 I3     ---   N''    Lower state N level
  39- 49 F11.5  ---   Obs    ? Observed transition position (2)
  51- 60 F10.4  ---   Cal    Position calculated by PGOPHER (3)
  62- 69 F8.5   ---   Res    ? Observed - calculated position
  71- 80 F10.4  ---   E''    Lower state energy calculated by PGOPHER (4)
  82- 93 E12.6  ---   A      Einstein A value
  95-106 E12.6  ---   f      The f-value
 108-118 A11    ---   Des    Transition description
--------------------------------------------------------------------------------
Note (1): In the A state, F=1 refers to omega=0.5, and F=2 to omega=1.5. In 
          the B and X states, F=1 refers to e parity, and F=2 to f parity.
Note (2): Blank if line not observed in any study.
Note (3): Using constants published in ????
Note (4): Relative to v''=0 band origin.
--------------------------------------------------------------------------------
A X  0  0   1.5   2.5 1 1 e e       2  9082.98330  9082.9702  0.01313    11.3536  7.685729E+3  9.310961E-5 pP1(2.5)
A X  0  0   2.5   3.5 1 1 e e       3  9079.91520  9079.8980  0.01723    22.7030  1.152579E+4  1.571904E-4 pP1(3.5)
A X  0  0   3.5   4.5 1 1 e e       4  9076.35990  9076.3587  0.00119    37.8337  1.405716E+4  2.046548E-4 pP1(4.5)
A X  0  0   4.5   5.5 1 1 e e       5  9072.35580  9072.3561 -0.00032    56.7451  1.596585E+4  2.423418E-4 pP1(5.5)
A X  0  0   5.5   6.5 1 1 e e       6  9067.89420  9067.8947 -0.00046    79.4362  1.751230E+4  2.736789E-4 pP1(6.5)
"""



"""
HITRAN Database [Gordon2016] downloader

References:

[Gordon2016] I.E. Gordon, L.S. Rothman, C. Hill, R.V. Kochanov, Y. Tan, P.F. Bernath, M. Birk,
    V. Boudon, A. Campargue, K.V. Chance, B.J. Drouin, J.-M. Flaud, R.R. Gamache, J.T. Hodges,
    D. Jacquemart, V.I. Perevalov, A. Perrin, K.P. Shine, M.-A.H. Smith, J. Tennyson, G.C. Toon,
    H. Tran, V.G. Tyuterev, A. Barbe, A.G. Császár, V.M. Devi, T. Furtenbacher, J.J. Harrison,
    J.-M. Hartmann, A. Jolly, T.J. Johnson, T. Karman, I. Kleiner, A.A. Kyuberis, J. Loos,
    O.M. Lyulin, S.T. Massie, S.N. Mikhailenko, N. Moazzen-Ahmadi, H.S.P. Müller, O.V. Naumenko,
    A.V. Nikitin, O.L. Polyansky, M. Rey, M. Rotger, S.W. Sharpe, K. Sung, E. Starikova,
    S.A. Tashkun, J. Vander Auwera, G. Wagner, J. Wilzewski, P. Wcisło, S. Yu, E.J. Zak,
    The HITRAN2016 Molecular Spectroscopic Database, J. Quant. Spectrosc. Radiat. Transf. (2017).
    doi:10.1016/j.jqsrt.2017.06.038.
"""

from robobrowser import RoboBrowser
import re


__all__ = ["get_hitran_molecules", "get_hitran_isotopologues"]


def get_hitran_molecules():
    """
    Accesses http://hitran.org/lbl/# and reads its table
    Returns: tuple: table (list of lists), header (list of strings)
    """

    data, header = [], []


    browser = RoboBrowser(history=True, parser="lxml")
    browser.open("http://hitran.org/lbl/#")

    table = browser.find("table")

    hh = table.find_all("th")
    for h in hh:
        # Skips cells whose class starts with "meta" (they are not of interest)
        cls = h.get("class")
        if isinstance(cls, list) and cls[0].startswith("meta"):
            continue
        header.append(h.text)

    rr = table.find_all("tr")
    for r in rr:
        dd = r.find_all("td")
        if len(dd) == 0:
            continue
        row = []
        data.append(row)
        for d in dd:
            # Skips cells whose class starts with "meta" (they are not of interest)
            cls = d.get("class")
            if isinstance(cls, list) and cls[0].startswith("meta"):
                continue
            row.append(d.text)

    return data, header


def get_hitran_isotopologues(id_molecule):
    """
    Accesses http://hitran.org/lbl/2 and reads its table

    Args:
        id_molecule: molecule ID as in HITRAN database (use get_hitran_molecules() to get
                     this information)

    Returns: tuple: table (list of lists), header (list of strings)
    """

    data, header = [], []


    browser = RoboBrowser(history=True, parser="lxml")
    browser.open("http://hitran.org/lbl/2?{}=on".format(id_molecule))

    table = browser.find("table")

    hh = table.find_all("th")
    for h in hh:
        # Skips cells whose class starts with "meta" (they are not of interest)
        cls = h.get("class")
        if isinstance(cls, list) and cls[0].startswith("meta"):
            continue
        header.append(h.text)

    rr = table.find_all("tr")
    for r in rr:
        dd = r.find_all("td")
        if len(dd) == 0:
            continue
        row = []
        data.append(row)
        for i, d in enumerate(dd):
            # Skips cells whose class starts with "meta" (they are not of interest)
            cls = d.get("class")
            if isinstance(cls, list) and cls[0].startswith("meta"):
                continue

            if i == 1:
                # **Note** Assumes that "Formula" is the second field
                formula = re.match("<td>(.*)</td>", str(d)).group(1)
                text = _slugify_html_formula(formula)
            else:
                text = d.text
            row.append(text)

    return data, header
    # rows = table.find_all("tr")
    # n = 0
    # header = None
    # data = []  # list of lists
    # for row in rows:
    #     cols = row.find_all("td")
    #     flag_header = False
    #     if len(cols) == 0:
    #         if n > 0:
    #             continue
    #         # If no columns found, we assume it is the header row
    #         cols = row.find_all("th")
    #         header = [f_header(ele) for ele in cols]
    #     elif len(cols) == 13:
    #         row = [function(ele) for function, ele in zip(functions, cols)]
    #         data.append(row)
    #         n += 1
    #
    # # print(plain)
    # #
    # # print(rows_table)

    # return data, header, title


def _slugify_html_formula(formula):
    """Converts HTML formula to plain text (deals with HTML superscript and subscript tags)

    This function transforms superscript text into text between parentheses:
        '<sup>x</sup>' becomes '(x)'

    Also transforms subscript text into plain text:
       '<sub>x</sub>' becomes 'x'
    """

    formula = re.sub("<sup>(.+?)</sup>", lambda g: "({})".format(g.group(1)), formula)
    formula = re.sub("<sub>(.+?)</sub>", lambda g: "{}".format(g.group(1)), formula)

    return formula

if __name__ == "__main__":
    import tabulate
    data, header = get_hitran_molecules()
    print(tabulate.tabulate(data, header))

    data, header = get_hitran_isotopologues(4)
    print(tabulate.tabulate(data, header))



#!/usr/bin/env python
"""
Convolve spectrum with Gaussian profile.

This is the Python version of "nulbad", the PFANT convolution-with-Gaussian
utility. This script ``nulbad.py`` was created in order to support more input
formats.

On the other hand, it does not open file "main.dat" to get parameters as
Fortran ``nulbad`` does. The latter also has a re-sampling ability that
this one doesn't.

The resulting file, compared to Fortran ``nulbad``, is nearly alike (you may see
a difference around the 6th digit). This one has one more data point at each side.

Supported file formats
======================

    Format                         Input?   Output?  Notes
    -----------------------------  -------  -------- ---------------------
    ``pfant`` output               YES      no    
    XY (two-column "lambda-flux")  YES      YES      default output format
    FITS                           YES      YES

Examples
========

  1) Create file 'flux.norm.pynulbad.0.08':

     nulbad.py --fwhm 0.08 flux.norm

  2) Create file 'flux.norm.pynulbad.0.08.fits':

     nulbad.py --fwhm 0.08 -f fits flux.norm

"""

import argparse
import logging
import sys
import a99
import f311
import pyfant
import f311
import numpy as np

a99.logging_level = logging.INFO

DEF_FN_CV = '<fn_flux>.<fwhm>'

OCLSS = {"xy": f311.FileSpectrumXY, "fits": f311.FileSpectrumFits}


def save(args, sp):

    f = OCLSS[args.format]()
    f.spectrum = sp
    f.save_as(args.fn_cv)

    print("Successfully created file '{0!s}'".format(args.fn_cv))


def get_gaussian(fwhm, pas):
    """Calculate Gaussian curve

    Args:
        fwhm
        pas: delta-lambda of curve

    Returns: 
        y: numpy vector containing calculated Gaussian curve
    """

    # Relation between the standard deviation and FWHM
    # (https://en.wikipedia.org/wiki/Full_width_at_half_maximum)
    sigma = fwhm/(2*np.sqrt(2*np.log(2)))

    NUM_SIGMAS = 3*1.4142
    num_points = 2*np.ceil(NUM_SIGMAS*sigma/pas)+1  # note this is an odd number
    offset = pas*((num_points-1)/2)

    x = np.arange(num_points, dtype=float)*pas-offset

    k1 = 1/(sigma*np.sqrt(2*np.pi))
    k2 = 2*sigma**2
    y = k1*np.exp(-x**2/k2)*pas

    return y

def convolve(args, sp):
    """Convolve spectrum with Gaussian profile and return new spectrum.

    The new spectrum has a narrower range"""

    sp2 = f311.Spectrum()

    h = get_gaussian(args.fwhm, sp.delta_lambda)
    num_points = len(h)

    sp2.y = np.convolve(sp.y, h, 'valid')
    off = int((num_points-1)/2)  # number of invalid points at each edge
    sp2.x = sp.x[off:-off]

    assert(len(sp2.x) == len(sp2.y))

    return sp2

def main(args):    
    sp = f311.load_spectrum(args.fn_flux)
    if not sp:
        print("File '{0!s}' not recognized as a spectrum file.".format(args.fn_flux))
        sys.exit()

    sp2 = convolve(args, sp)

    save(args, sp2)






if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
 
    parser.add_argument('--fwhm', type=float, nargs='?', default=0.12, 
        help="full width at half maximum (FWHM) of Gaussian curve")
    parser.add_argument('-f', '--format', type=str, nargs='?', default='xy',
        choices=["xy", "fits"], help="Output format")
    parser.add_argument('fn_flux', type=str,  
        help='input spectral filename')
    parser.add_argument('fn_cv', type=str, help='output file name', nargs='?', default=DEF_FN_CV)

    args = parser.parse_args()

    if args.fn_cv == DEF_FN_CV:
        more = ".fits" if args.format == "fits" else ""
        args.fn_cv = "{}.pynulbad.{:.3f}{}".format(args.fn_flux, args.fwhm, more)

    print("fn_flux (input filename) =", args.fn_flux)
    print("fn_cv (output filename) =", args.fn_cv)
    print("fwhm =", args.fwhm)
    print("output format = ", args.format)

    main(args)

 



"""Widget to edit a Conf object."""

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from ._shared import *
import pyfant
import f311


__all__ = ["WOptionsEditor"]

_FROM_MAIN = ' (read from main configuration file)'
_EXE_NAMES = {"i": "innewmarcs", "h": "hydro2", "p": "pfant", "n": "nulbad"}

IHPN = "ihpn"
NIHPN = len(IHPN)


@a99.froze_it
class _Option(a99.AttrsPart):
    attrs = ["name", "argname", "descr"]

    def __init__(self):
        a99.AttrsPart.__init__(self)
        self.checkbox = None
        self.label = None
        self.edit = None
        self.name = None
        self.ihpn = None
        self.argname = None
        self.default = None
        self.short_descr = None
        self.long_descr = None
        # Whether or not the option overrides a value in main.dat
        self.flag_main = False
        # Whether or not the option is considered to be an option for developers
        # (to aid debuggin etc)
        self.flag_devel = False
        self.color = a99.COLOR_DESCR
        # other widgets to be shown/hidden (besides .checkbox, .label, .edit)
        self.other_widgets = []
        # Type (e.g. float, str). Only required for options whose default is None
        self.type = None
        self.flag_never_used = True
        self.possible_values = None
        self.min = None
        self.max = None

    def __clear_edit(self):
        if isinstance(self.edit, QCheckBox):
            self.edit.setChecked(False)
        else:
            self.edit.setText("")

    def __update_edit(self, value):
        if value is None:
            if isinstance(self.edit, QCheckBox):
                self.edit.setChecked(False)
            else:
                self.edit.setText("")
        else:
            if isinstance(self.edit, QCheckBox):
                self.edit.setChecked(value == True)
            else:
                self.edit.setText(str(value))

    def update_edit_with_default(self):
        self.__update_edit(self.default)

    def update_gui(self, options, flag_reset=False):
        """Updates control and in-use-checkbox.

        Args:
          options: FileOptions instance
          flag_reset: resets controls to default value if their value in the
                              FileOptions object is None
        """
        assert isinstance(options, pyfant.FileOptions)
        attr = options.__getattribute__(self.name)
        flag_check = attr is not None
        self.checkbox.setChecked(flag_check)
        if self.flag_never_used and flag_check:
            self.flag_never_used = False
        if flag_check:
            self.__update_edit(attr)
        elif flag_reset:
            self.update_edit_with_default()

    def get_label_text(self):
        return a99.enc_name_descr("--%s" % self.name, self.short_descr, self.color)

    def get_value(self):
        type_ = type(self.default) if self.default is not None else self.type
        w = self.edit
        if type_ == bool:
            value = w.isChecked()
        else:
            s = str(w.text())

            # validation
            if len(s) == 0:
                raise RuntimeError("empty")

            value = type_(s)

            # validation
            if self.possible_values is not None:
                if value not in self.possible_values:
                    raise RuntimeError("invalid value: \"%s\" -- possible values are: %s" %
                                       (str(value), '"' + (
                                       '"/"'.join([str(x) for x in self.possible_values])) + '"'))
            if self.min is not None and value < self.min:
                raise RuntimeError("value too small (minimum allowed: {})".format(self.min))
            if self.max is not None and value > self.max:
                raise RuntimeError("value too big (maximum allowed: {})".format(self.max))

        return value


class WOptionsEditor(a99.WEditor):
    """
    FileOptions editor widget.

    Args:
      parent=None
    """

    def __init__(self, parent=None):
        a99.WEditor.__init__(self, parent)

        # # Setup & accessible attributes

        # # Internal stuff that must not be accessed from outside

        # options map: list of _Option
        self.omap = []
        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False
        self.error_text = ""
        self.hiding_text = ""

        # # Central layout

        la = self.layout_editor
        a99.set_margin(la, 0)

        # ## Toolbar: checkboxes with executables
        l1 = self.layout_exes = QHBoxLayout()
        la.addLayout(l1)

        w = self.label_filter = QLabel("<b>Fil&ter:</b>")
        l1.addWidget(w)
        w = self.lineEdit_filter = QLineEdit()
        self.label_filter.setBuddy(w)
        w.textEdited.connect(self.on_filter)
        l1.addWidget(w)
        w.setFixedWidth(100)

        w = self.line12312 = QFrame()  # vertical line
        l1.addWidget(w)
        # w.setGeometry(QRect(240, 240, 3, 61))
        w.setFrameShape(QFrame.VLine)
        w.setFrameShadow(QFrame.Sunken)
        w.setFixedWidth(3)

        w0 = self.checkbox_i = QCheckBox("innewmarcs")
        w1 = self.checkbox_h = QCheckBox("hydro2")
        w2 = self.checkbox_p = QCheckBox("pfant")
        w3 = self.checkbox_n = QCheckBox("nulbad")
        ww = self.checkboxes_exes = [w0, w1, w2, w3]
        l1.addWidget(QLabel("<b>Show options for Fortran binaries:</b>"))
        for w in ww:
            l1.addWidget(w)
            w.setTristate(False)
            w.setChecked(True)
            w.stateChanged.connect(self.on_checkbox_exe_clicked)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # ## Toolbar: checkbox for main configuration file override
        l1 = self.layout23432 = QHBoxLayout()
        la.addLayout(l1)
        w = self.checkbox_main = QCheckBox("Show options that override main configuration file")
        w.setStyleSheet("QCheckBox {color: %s}" % COLOR_CONFIG)
        w.setChecked(True)
        w.stateChanged.connect(self.on_checkbox_main_clicked)
        w.setToolTip("Show options that, if set, will override values "
                     "that appear in the main configuration file.")
        l1.addWidget(w)
        w = self.checkbox_devel = QCheckBox("Show developer options")
        w.setChecked(False)
        w.stateChanged.connect(self.on_checkbox_devel_clicked)
        w.setToolTip("Show options that were created to aid software development and debugging.")
        l1.addWidget(w)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        b = QPushButton("Preview command line")
        b.clicked.connect(self.on_preview)
        l1.addWidget(b)

        # ## Splitter with scroll area and descripton+error area
        sp = self.splitter = QSplitter(Qt.Vertical)
        la.addWidget(sp)

        # ### Scroll area containing the form

        sa = self.scrollArea = QScrollArea()
        sp.addWidget(sa)
        sa.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        sa.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        # Widget that will be handled by the scrollable area
        w = self.scrollWidget = QWidget()
        sa.setWidget(self.scrollWidget)
        sa.setWidgetResizable(True)
        #        la.addWidget(w)

        # #### The form layout
        # This layout contains the form layout and a spacer item
        lw = QVBoxLayout()
        w.setLayout(lw)
        # Form layout
        lo = self.formLayout = QGridLayout()
        lw.addLayout(lo)
        lo.setVerticalSpacing(4)
        lo.setHorizontalSpacing(5)
        lw.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        # ##### The editing controls

        self.w_logging_level = QLineEdit()
        self.w_logging_console = QCheckBox()
        self.w_allow = QCheckBox()
        self.w_logging_file = QCheckBox()
        self.w_fn_logging = QLineEdit()
        self.w_fn_main = QLineEdit()
        self.w_fn_progress = QLineEdit()
        self.w_explain = QCheckBox()
        self.w_fn_modeles = QLineEdit()
        self.w_teff = QLineEdit()
        self.w_glog = QLineEdit()
        self.w_asalog = QLineEdit()
        self.w_fn_absoru2 = QLineEdit()
        self.w_fn_hmap = QLineEdit()
        self.w_llzero = QLineEdit()
        self.w_llfin = QLineEdit()
        self.w_fn_moddat = QLineEdit()
        self.w_modcode = QLineEdit()
        self.w_tirb = QLineEdit()
        self.w_ptdisk = QLineEdit()
        self.w_kik = QLineEdit()
        self.w_amores = QCheckBox()
        self.w_kq = QLineEdit()
        self.w_vvt = QLineEdit()
        self.w_zph = QLineEdit()
        self.w_interp = QLineEdit()
        self.w_fn_dissoc = QLineEdit()
        self.w_fn_partit = QLineEdit()
        self.w_fn_abonds = QLineEdit()
        self.w_fn_atoms = QLineEdit()
        self.w_fn_moo = QLineEdit()
        self.w_fn_modgrid = QLineEdit()
        self.w_fn_molecules = QLineEdit()
        self.w_fn_opa = QLineEdit()
        self.w_no_molecules = QCheckBox()
        self.w_no_atoms = QCheckBox()
        self.w_no_h = QCheckBox()
        self.w_opa = QCheckBox()
        self.w_abs = QCheckBox()
        self.w_sca = QCheckBox()
        self.w_absoru = QCheckBox()
        self.w_zinf = QLineEdit()
        self.w_pas = QLineEdit()
        self.w_aint = QLineEdit()
        self.w_flprefix = QLineEdit()
        self.w_fn_flux = QLineEdit()
        self.w_flam = QCheckBox()
        self.w_fn_cv = QCheckBox()
        self.w_pat = QLineEdit()
        self.w_convol = QCheckBox()
        self.w_fwhm = QLineEdit()

        # ##### The options map
        # (*) A few options have been commented out because they are probably
        # irrelevant nowadays, but may be shown again some time.

        o = self.__add_option(self.w_logging_level, 'ihpn', 'logging_level', 'debug',
                              'logging level',
                              'These are the available options:<ul>' +
                              '<li>debug' +
                              '<li>info' +
                              '<li>warning' +
                              '<li>error' +
                              '<li>critical' +
                              '<li>halt</ul>')
        o.flag_devel = True
        o.possible_values = ["debug", "info", "warning", "error", "critical", "halt"]
        o = self.__add_option(self.w_logging_console, 'ihpn', 'logging_console', True,
        'Log to console?',
        'Whether or not to log messages to standard output (usually the command console)')
        o.flag_devel = True
        o = self.__add_option(self.w_logging_file, 'ihpn', 'logging_file', False,
                              'Log to file?',
                              'Whether or not to log messages to log file ' +
                              '(specified by option --fn_logging)')
        o.flag_devel = True
        o = self.__add_option(self.w_explain, 'ihpn', 'explain', False,
        'Save additional debugging information?',
        'This flag informs the Fortran code to save additional information in file explain.txt ' +
        '(debugging purposes; output varies, or flag may be ignored)')
        o.flag_devel = True
        o = self.__add_option(self.w_fn_logging, 'ihpn', 'fn_logging', None,
        'log filename',
        'default: <executable name>_dump.log, <i>e.g.</i>, <b>pfant_dump.log</b>')
        o.type = str
        o.flag_devel = True
        self.__add_option(self.w_fn_main, 'ihpn', 'fn_main', pyfant.FileMain.default_filename,
                          'input file name: main configuration',
                          'Contains star parameters and additional software configuration.<br><br>'
                          '<b>Attention</b>: the following command-line options, if used, will '
                          'override values in the main configuration file: '
                          '--llzero, --llfin, --pas, --flprefix, --fwhm.')
        # (*) self.__add_option(self.w_fn_progress, 'ihpn', 'fn_progress', 'file name',
        # 'progress.txt',
        # (*)  'output file name - progress indicator')

        #
        # innewmarcs, hydro2, pfant
        #
        self.__add_option(self.w_fn_modeles, 'ihp', 'fn_modeles', pyfant.FileModBin.default_filename,
                          'atmospheric model file name',
                          'This is a binary file containing information about atmospheric model. '
                          'This file is created by innewmarcs.')

        #
        # innewmarcs, pfant
        #
        self.__add_option(self.w_opa, 'ip', 'opa', True,
        'Use MARCS opacities?',
        'Whether or not to include MARCS opacity coefficients (absorption and scattering) '
        'in the continuum.')
        # the following option is a pfant-only option but it is better to put it here
        o = self.__add_option(self.w_absoru, 'p', 'absoru', True,
        'Use absoru() kappa?',
        'Whether or not to include coefficients calculated by subroutine absoru() '
        'in the continuum.')
        o.flag_devel = False  # True
        self.__add_option(self.w_fn_opa, 'ip', 'fn_opa', pyfant.FileOpa.default_filename,
        'Opacities filename',
        'This is a text file in the MARCS ".opa" format. This file can be generated through '
        'interpolation using <em>innewmarcs</em>.'
        '<p>Files downloaded directly from the MARCS '
        'website should also work, although it is expected that you will use <em>innewmarcs</em>'
        'to interpolate this opacities file based on a grid of models.')

        #
        # innewmarcs-only
        #
        # (*) self.__add_option(self.w_modcode, 'i', 'modcode', 'string up to 25 characters',
        # 'NoName',
        # (*)  '"Model name"')
        self.__add_option(self.w_fn_modgrid, 'i', 'fn_modgrid', "grid.mod",
        'atmospheric model grid (<b>without opacities</b>)',
        'This is a binary file containing a grid of atmospheric models for interpolation.'
        '<p>Whether this file or the one specified by <em>--fn_moo</em> will be used '
        'will depend on the <em>--opa</em>option.')
        self.__add_option(self.w_fn_moo, 'i', 'fn_moo', pyfant.FileMoo.default_filename,
        'atmospheric model grid (<b>with opacities</b>)',
        'This is a binary file containing a grid of atmospheric models for interpolation, '
        'opacities included.'
        '<p>Whether this file or the one specified by <em>--fn_modgrid</em> will be used '
        'will depend on the <em>--opa</em>option.')
        self.__add_option(self.w_allow, 'i', 'allow', False,
        'Allow point out of atmospheric model grid?',
        'If this option is disabled, target '
        ' (glog, teff, asalog) point for interpolation must lie inside the 3D '
        'atmospheric model grid.')

        #
        # hydro2-only
        #
        self.__add_option(self.w_amores, 'h', 'amores', True,
                          'AMOrtissement de RESonnance?', '')
        o = self.__add_option(self.w_kq, 'h', 'kq', 1,
                              'theorie',
                              '<ul>'
                              '<li>0: THEORIE DE GRIEM' +
                              '<li>1: THEORIE QUASISTATIQUE</ul>')
        o.possible_values = [0, 1]

        o = self.__add_option(self.w_zph, 'h', 'zph', 12.,
        'hydrogen-reference-abundance',
        'abundance d\'H pour laquelle sont donn&eacute;es les abondances metalliques')
        o.min = 0.
        o.max = 24.  # actually I would like to remove this option

        #
        # hydro2, pfant
        #
        o = self.__add_option(self.w_llzero, 'hp', 'llzero', 6000.,
                              "lower boundary of calculation interval (&Aring;)",
                              'default: &lt;main_llzero&gt; ' + _FROM_MAIN)
        o.flag_main = True
        o.color = COLOR_CONFIG
        o.type = float
        o = self.__add_option(self.w_llfin, 'hp', 'llfin', 6200.,
                              'upper boundary of calculation interval (&Aring;)',
                              'default: &lt;main_llfin&gt; ' + _FROM_MAIN)
        o.flag_main = True
        o.color = COLOR_CONFIG
        o.type = float
        self.__add_option(self.w_fn_absoru2, 'hp', 'fn_absoru2', pyfant.FileAbsoru2.default_filename,
                          'input file name - absoru2',
                          'This file contains physical data for pfant and hydro2 "absoru" module')
        self.__add_option(self.w_fn_hmap, 'hp', 'fn_hmap', pyfant.FileHmap.default_filename,
                          'input file name - hydrogen lines data',
                          'Contains a table with<pre>' +
                          'filename, niv inf, niv sup, central lambda, kiex, c1</pre>'
                          'for each hydrogen line.')
        o = self.__add_option(self.w_kik, 'hp', 'kik', 0,
        'option affecting the flux integration',
        '<ul>' +
        '<li>0: integration using 6/7 points depending on <em>ptdisk</em> '
        'parameter in main configuration file;' +
        '<li>1: 26-point integration</ul>')
        o.possible_values = [0, 1]

        #
        # pfant-only
        #
        o = self.__add_option(self.w_aint, 'p', 'aint', 10,
                              'interval length per iteration (&Aring;)',
                              'default: &lt;main_aint&gt; ' + _FROM_MAIN)
        o.flag_devel = True
        o = self.__add_option(self.w_interp, 'p', 'interp', 1,
                              'interpolation for subroutine turbul()',
                              'interpolation type for subroutine turbul()<ul>' +
                              '<li>1: linear;' +
                              '<li>2: parabolic</ul>')
        o.possible_values = [1, 2]
        # > @todo Find names for each file and update options help
        self.__add_option(self.w_fn_dissoc, 'p', 'fn_dissoc', pyfant.FileDissoc.default_filename,
                          'input file name - dissociative equilibrium', '')
        self.__add_option(self.w_fn_partit, 'p', 'fn_partit', pyfant.FilePartit.default_filename,
                          'input file name - partition functions', '')
        self.__add_option(self.w_fn_abonds, 'p', 'fn_abonds', pyfant.FileAbonds.default_filename,
                          'input file name - atomic abundances', '')
        self.__add_option(self.w_fn_atoms, 'p', 'fn_atoms', pyfant.FileAtoms.default_filename,
                          'input file name - atomic lines', '')
        self.__add_option(self.w_fn_molecules, 'p', 'fn_molecules',
                          pyfant.FileMolecules.default_filename,
                          'input file name - molecular lines', '')
        self.__add_option(self.w_no_molecules, 'p', 'no_molecules', False,
                          'Skip molecules?',
                          'If set, skips the calculation of molecular lines.')
        self.__add_option(self.w_no_atoms, 'p', 'no_atoms', False,
                          'Skip atomic lines?',
                          'If set, skips the calculation of atomic lines (except hydrogen).')
        self.__add_option(self.w_no_h, 'p', 'no_h', False,
                          'Skip hydrogen lines?',
                          'If set, skips the calculation of hydrogen lines.')
        o = self.__add_option(self.w_zinf, 'p', 'zinf', 0.5,
        '(zinf per-line in dfile:atoms)',
        'This is the distance from center of line to consider in atomic line calculation.<br><br>' +
        'If this option is used, it will override all the zinf defined ' +
        'in the atomic lines file with a single value.')
        o.flag_devel = True
        o = self.__add_option(self.w_pas, 'p', 'pas', 0.02,
                              'calculation delta-lambda (&Aring;)',
                              'default: &lt;main_pas&gt; ' + _FROM_MAIN)
        o.flag_main = True
        o.color = COLOR_CONFIG
        o = self.__add_option(self.w_abs, 'p', 'abs', False,
        'Use MARCS absorption coefficients?',
        'Whether or not to include MARCS <b>absorption</b> coefficients in the continuum'
        '<p>This option only has effect if <em>--opa<em> is True.')
        o.flag_devel = False
        o = self.__add_option(self.w_sca, 'p', 'sca', True,
        'Use MARCS scattering coefficients?',
        'Whether or not to include MARCS <b>scattering</b> coefficients in the continuum'
        '<p>This option only has effect if <em>--opa<em> is True.')
        o.flag_devel = False

        #
        # pfant, nulbad
        #
        o = self.__add_option(self.w_flprefix, 'pn', 'flprefix', 'flux',
                              'pfant output - prefix for flux output files',
                              'Three files will be created based on this prefix:<ul>' +
                              '<li><flprefix>.spec: un-normalized spectrum' +
                              '<li><flprefix>.cont: continuum' +
                              '<li><flprefix>.norm: normalized spectrum</ul><br>'
                              'default: &lt;main_flprefix&gt; ' + _FROM_MAIN)
        o.flag_main = True
        o.color = COLOR_CONFIG

        #
        # nulbad-only
        #
        self.__add_option(self.w_fn_flux, 'n', 'fn_flux', 'flux.norm',
                          'nulbad input - flux file name',
                          'default: &lt;main_flprefix>.norm ' + _FROM_MAIN)
        o.flag_main = True
        o.color = COLOR_CONFIG
        self.__add_option(self.w_flam, 'n', 'flam', False,
                          'Perform Fnu-to-FLambda transformation?',
                          'If True, Fnu-to-FLambda transformation is done before the convolution')
        self.__add_option(self.w_fn_cv, 'n', 'fn_cv', True,
                          'nulbad output -- convolved spectrum file name',
                          'default: &lt;flux file name&gt;.nulbad.&lt;fwhm&gt;')
        self.__add_option(self.w_pat, 'n', 'pat', 0.02,
                          'wavelength delta-lambda of nulbad output spectrum (&Aring;)',
                          'default: same as delta-lambda of input spectrum')
        self.__add_option(self.w_convol, 'n', 'convol', True,
                          'Apply convolution?', '')
        o = self.__add_option(self.w_fwhm, 'n', 'fwhm', 0.12,
                              'full-width-half-maximum of Gaussian function',
                              'default: &lt;main_fwhm&gt; ' + _FROM_MAIN + ')')
        o.flag_main = True
        o.color = COLOR_CONFIG

        for j, letter in enumerate(IHPN):
            label = QLabel("<b>%s</b>" % letter)
            label.setToolTip("Option is used by %s?" % _EXE_NAMES[letter])
            lo.addWidget(label, 0, 3 + j, Qt.AlignCenter)
        label = QLabel("<b>in use?</b>")
        label.setToolTip("Option will appear in command line?")
        lo.addWidget(label, 0, 2)

        i = 1  # index of next row in layout
        for option in self.omap:
            try:
                for j, letter in enumerate(IHPN):
                    # unicode is for a "v"-like check mark
                    flag_checked = letter in option.ihpn
                    # label = QLabel(QString(chr(10003)) if flag_checked else "")
                    label = QLabel("\u2713" if flag_checked else "")
                    if flag_checked:
                        label.setToolTip("This option is used by %s." % _EXE_NAMES[letter])

                    option.other_widgets.append(label)
                    lo.addWidget(label, i, 3 + j, Qt.AlignCenter)
                label = option.label = \
                    QLabel(option.get_label_text())
                label.setAlignment(Qt.AlignRight)
                checkbox = option.checkbox = QCheckBox()
                checkbox.installEventFilter(self)
                checkbox.stateChanged.connect(self.on_in_use_checkbox_clicked)
                edit = option.edit
                edit.installEventFilter(self)

                lo.addWidget(label, i, 0)
                lo.addWidget(edit, i, 1)
                lo.addWidget(checkbox, i, 2, Qt.AlignCenter)
                i += 1
                for w in [option.label, option.edit]:
                    w.setToolTip(option.long_descr)

                if isinstance(edit, QCheckBox):
                    edit.stateChanged.connect(self.on_edited)
                else:
                    edit.textEdited.connect(self.on_edited)

                    # todo No consensus yet if it is better to show the default values
                    # option.update_edit_with_default()
            except:
                a99.get_python_logger().exception("Processing option '%s'" % option.name)
                raise
        self.__update_from_data(pyfant.FileOptions(), True)

        # ### Second widget of splitter
        # layout containing description area and a error label
        wlu = QWidget()
        lu = QVBoxLayout(wlu)
        a99.set_margin(lu, 0)
        lu.setSpacing(4)
        x = self.textEditDescr = QTextEdit(self)
        x.setReadOnly(True)
        # x.setGeometry(0, 0, 100, 0)
        # x.setWordWrap(True)
        x.setStyleSheet("QTextEdit {color: %s}" % a99.COLOR_DESCR)
        lu.addWidget(x)

        x = self.labelError = QLabel(self)
        x.setStyleSheet("QLabel {color: %s}" % a99.COLOR_ERROR)
        lu.addWidget(x)

        sp.addWidget(wlu)

        # ### Adjust splitter proportion
        sp.setStretchFactor(0, 8)
        sp.setStretchFactor(1, 2)

        self.__update_gui_visible_options()
        a99.style_checkboxes(self)
        self.setEnabled(False)  # disabled until load() is called
        self.flag_process_changes = True

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def _do_load(self, x):
        assert isinstance(x, pyfant.FileOptions)
        self._f = x
        self.__update_from_data(flag_reset=True)
        # this is called to perform file validation upon loading
        self.__update_data()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""

        # todo not working, dunno why but it is called but the first field does not acquire focus.

        self.omap[0].edit.setFocus()

    def eventFilter(self, obj_focused, event):
        if event.type() == QEvent.FocusIn:
            option = self.__find_option_by_widget(obj_focused)
            if option:
                text = "%s<br><br>%s" % \
                       (a99.enc_name(option.name.replace("&", ""), option.color),
                        option.long_descr)
                self.__set_descr_text(text)

        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_edited(self):
        if not self.flag_process_changes:
            return
        # self._update_file_main()
        option = self.__find_option_by_edit(self.sender())
        self.flag_process_changes = False
        option.checkbox.setChecked(True)
        self.flag_process_changes = True
        option.flag_never_used = False
        self.__update_data()
        self.changed.emit()

    def on_in_use_checkbox_clicked(self):
        if not self.flag_process_changes:
            return
        self.__update_data()
        self.changed.emit()

    def on_checkbox_exe_clicked(self):
        self.__update_gui_visible_options()

    def on_checkbox_main_clicked(self):
        self.__update_gui_visible_options()

    def on_checkbox_devel_clicked(self):
        self.__update_gui_visible_options()

    def on_preview(self):
        args = self.f.get_args()
        print(args)
        line = "fortran-binary-xxxx " + (" ".join(args))
        w = f311.XText(self, line, "Command line")
        w.show()

    def on_filter(self):
        if not self.flag_process_changes:
            return
        self.__update_gui_visible_options()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear

    def __add_option(self, widget, ihpn, name, default, short_descr, long_descr):
        o = _Option()
        w = o.checkbox = QCheckBox()
        w.setTristate(False)
        w.stateChanged.connect(self.on_in_use_checkbox_clicked)

        o.edit = widget
        o.ihpn = ihpn
        o.name = name
        o.default = default
        o.short_descr = short_descr
        o.long_descr = long_descr
        self.omap.append(o)
        return o

    def __find_option_by_widget(self, widget):
        if isinstance(widget, QCheckBox):
            ret = self.__find_option_by_in_use_checkbox(widget)
            if not ret:
                ret = self.__find_option_by_edit(widget)
        else:
            ret = self.__find_option_by_edit(widget)
        return ret

    def __find_option_by_in_use_checkbox(self, checkbox):
        ret = None
        for option in self.omap:
            if option.checkbox == checkbox:
                ret = option
                break
        return ret

    def __find_option_by_edit(self, widget):
        ret = None
        for option in self.omap:
            if option.edit == widget:
                ret = option
                break
        return ret

    def __update_from_data(self, data=None, flag_reset=False):
        """Updates the edits and in-use-checkboxes.

        data=self.f: FileOptions instance
        flag_reset: options are loaded as if for the first time into GUI
        """
        self.flag_process_changes = False
        try:
            if data is None:
                data = self.f
            for option in self.omap:
                option.update_gui(data, flag_reset)
        finally:
            self.flag_process_changes = True

    def __update_data(self):
        emsg, flag_error = "", False
        ss = ""
        try:
            for option in self.omap:
                ss = option.name
                if not option.checkbox.isChecked():
                    value = None
                else:
                    value = option.get_value()
                self.f.__setattr__(option.name, value)

            ss = ""
            if self.f.llzero is not None and self.f.llfin is not None:
                if self.f.llzero >= self.f.llfin:
                    raise RuntimeError("llzero must be lower than llfin")

        except Exception as E:
            flag_error = True
            if ss:
                emsg = "Option <em>--%s</em>: %s" % (ss, str(E))
            else:
                emsg = str(E)
            a99.get_python_logger().exception("Updating Options object")

        self._flag_valid = not flag_error
        self.__set_error_text(emsg)

    def __update_gui_visible_options(self):
        LETTERS = "ihpn"
        ll = []
        for i, (checkbox, letter) in enumerate(zip(self.checkboxes_exes, LETTERS)):
            if checkbox.isChecked():
                ll.append(letter)
        flag_main_goes = self.checkbox_main.isChecked()
        flag_devel_goes = self.checkbox_devel.isChecked()
        hidden_set_count = 0  # whether there will be set options hidden
        s_filter = str(self.lineEdit_filter.text()).upper()
        flag_filter = len(s_filter) > 0
        for option in self.omap:
            flag_visible = (flag_main_goes or not option.flag_main) and \
                           (flag_devel_goes or not option.flag_devel) and \
                           any([letter in option.ihpn for letter in ll])

            if flag_visible and flag_filter:
                flag_visible = s_filter in option.name.upper() or \
                               s_filter in option.short_descr.upper() or \
                               s_filter in option.long_descr.upper()

            option.checkbox.setVisible(flag_visible)
            option.label.setVisible(flag_visible)
            option.edit.setVisible(flag_visible)
            for w in option.other_widgets:
                w.setVisible(flag_visible)
            if not flag_visible and option.checkbox.isChecked():
                hidden_set_count += 1

        t = ""
        if hidden_set_count > 0:
            t = "Hiding %d option%s in use." % \
                (hidden_set_count, "" if hidden_set_count == 1 else "s")
        self.__set_hiding_text(t)

    def __set_error_text(self, x):
        """Sets text of labelError."""
        self.error_text = x
        self.__update_error_label()

    def __set_hiding_text(self, x):
        """Sets text of labelError."""
        self.hiding_text = x
        self.__update_error_label()

    def __set_descr_text(self, x):
        """Sets text of labelDescr."""
        self.textEditDescr.setText(x)

    def __update_error_label(self):
        l = []
        if self.error_text:
            l.append('<span style="color: %s"><b>Error</b>: %s</span>' %
                     (a99.COLOR_ERROR, self.error_text))
        if self.hiding_text:
            l.append('<span style="color: %s"><b>Warning</b>: %s</span>' %
                     (a99.COLOR_WARNING, self.hiding_text))
        s = "; ".join(l)
        self.labelError.setText(s)



#!/usr/bin/env python

"""
Merges several atmospheric models into a single file (_i.e._, the "grid")

"Collects" several files in current directory and creates a single file
containing atmospheric model grid.

Working modes (option "-m"):
 "opa" (default mode): looks for MARCS[1] ".mod" and ".opa" text file pairs and
                       creates a *big* binary file containing *all* model
                       information including opacities.
                       Output will be in ".moo" format.

 "modtxt": looks for MARCS ".mod" text files only. Resulting grid will not contain
           opacity information.
           Output will be in binary ".mod" format.


 "modbin": looks for binary-format ".mod" files. Resulting grid will not contain
           opacity information.
           Output will be in binary ".mod" format.

References:
  [1] http://marcs.astro.uu.se/

.
.
.
"""

import argparse
import logging
import glob
import os
import sys
import a99
import pyfant


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    parser.add_argument('--pattern', type=str, help='file name pattern (with wildcards)',
     nargs="?", default="*.mod")
    parser.add_argument('-m', '--mode', type=str, nargs="?", default="opa",
     choices=["opa", "modtxt", "modbin"],
     help='working mode (see description above)')
    VDOM = "\"grid.moo\" or \"grid.mod\", depending on mode"
    parser.add_argument('fn_output', type=str, help='output file name', nargs="?",
     default=VDOM)

    args = parser.parse_args()

    logger = a99.get_python_logger()

    if args.fn_output == VDOM:
        args.fn_output = "grid.moo" if args.mode == "opa" else "grid.mod"
        logger.info("Setting output filename to '{0!s}'".format(args.fn_output))

    filenames = glob.glob("./"+args.pattern)
    n = len(filenames)
    print("{0:d} file{1!s} matching pattern '{2!s}'".format(n, "s" if n != 1 else "", args.pattern))

    records = []
    for filename in filenames:
        if args.mode == "opa":
            name = os.path.splitext(os.path.basename(filename))[0]
            print("Considering files '{0!s}'+('.mod', '.opa') ...".format(name))
            try:
                f = pyfant.FileModTxt()
                f.load(filename)
                g = pyfant.FileOpa()
                g.load(name+".opa")
                r = pyfant.MooRecord()
                r.from_marcs_files(f, g)
                records.append(r)
            except:
                logger.exception("Error loading file '{0!s}', skipping...".format(filename))
        else:
            nameext = os.path.basename(filename)
            print("Considering file '{0!s}'+('.mod', '.opa') ...".format(nameext))
            try:
                if args.mode == "modtxt":
                    f = pyfant.FileModTxt()
                    f.load(filename)
                    records.append(f.record)
                else:
                    f = pyfant.FileModBin()
                    f.load(filename)
                    records.extend(f.records)
            except:
                logger.exception("Error loading file '{0!s}', skipping...".format(filename))


    if len(records) == 0:
        print("No valid models found, nothing to save.")
        sys.exit()

    records.sort(key=lambda r: r.asalog*1e10+r.teff*100+r.glog)

    if args.mode == "opa":
        g = pyfant.FileMoo()
    else:
        g = pyfant.FileModBin()
    g.records = records
    g.save_as(args.fn_output)

    print("Successfully created file '{0!s}'".format(args.fn_output))
